#define _GNU_SOURCE

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>

static const char *fs_check_path(lua_State *ctx, int i, size_t *psize) {
	size_t n;
	const char *path = luaL_checklstring(ctx, i, &n);

	size_t real_len = strlen(path);
	if (n != real_len) {
		luaL_error(ctx, "path may not contain \\0");
		return NULL;
	}

	if (psize) *psize = n;

	return path;
}

static void free_exec(char **env, size_t env_n) {
	for (size_t i = 0; i < env_n; i++) free(env[i]);
	if (env) free(env);
}

static int lib_close(lua_State *ctx) {
	luaL_Stream *stream = luaL_checkudata(ctx, 1, LUA_FILEHANDLE);
	fflush(stream->f);
	int stream_fd = fileno(stream->f);
	close(stream_fd);
}

static int lib_close_in(lua_State *ctx) {
	luaL_Stream *stream = luaL_checkudata(ctx, 1, LUA_FILEHANDLE);
	int stream_fd = fileno(stream->f);
	dup2(STDIN_FILENO, stream_fd);
}
static int lib_close_out(lua_State *ctx) {
	luaL_Stream *stream = luaL_checkudata(ctx, 1, LUA_FILEHANDLE);
	int stream_fd = fileno(stream->f);
	dup2(STDOUT_FILENO, stream_fd);
}
static int lib_close_err(lua_State *ctx) {
	luaL_Stream *stream = luaL_checkudata(ctx, 1, LUA_FILEHANDLE);
	int stream_fd = fileno(stream->f);
	dup2(STDERR_FILENO, stream_fd);
}

static int lib_reap(lua_State *ctx) {
	pid_t pid = lua_tointeger(ctx, lua_upvalueindex(1));
	if (pid < 0) return luaL_error(ctx, "Tried to reap child twice");

	lua_pushinteger(ctx, -1);
	lua_replace(ctx, lua_upvalueindex(1));

	int stat;
	waitpid(pid, &stat, 0);

	if (!WIFEXITED(stat)) {
		lua_pushinteger(ctx, -1);
		return 1;
	}
	else {
		lua_pushinteger(ctx, WEXITSTATUS(stat));
		return 1;
	}
}

static void close_pipes(int in[2], int out[2], int err[2]) {
	if (in[0] >= 0) close(in[0]);
	if (in[1] >= 0) close(in[1]);
	if (out[0] >= 0) close(out[0]);
	if (out[1] >= 0) close(out[1]);
	if (err[0] >= 0) close(err[0]);
	if (err[1] >= 0) close(err[1]);
}

static int lib_exec(lua_State *ctx) {
	luaL_checktype(ctx, 1, LUA_TTABLE);

	size_t n = lua_rawlen(ctx, 1);
	if (n <= 0) return luaL_error(ctx, "Table argument must have at least one array element");

	const char *args[n + 1];
	char **env;
	size_t env_cap = 0, env_n = 0;

	args[n] = NULL;

	for (size_t i = 0; i < n; i++) {
		lua_geti(ctx, 1, i + 1);
		if (!lua_isstring(ctx, -1)) return luaL_error(ctx, "Table argument must consist only of strings");

		args[i] = lua_tostring(ctx, -1);
		lua_pop(ctx, 1);
	}

	if (lua_getfield(ctx, 1, "env")) {
		lua_pushnil(ctx);
		while (lua_next(ctx, 1)) {
			if (lua_type(ctx, -2) == LUA_TSTRING) {
				if (!lua_isstring(ctx, -1)) {
					free_exec(env, env_n);
					return luaL_error(ctx, "Table argument must consist only of strings");
				}

				size_t prev_cap = env_cap;
				while (env_n >= env_cap) {
					if (env_cap == 0) env_cap = 16;
					else env_cap *= 2;
				}

				if (prev_cap) env = realloc(env, sizeof *env * env_cap);
				else env = malloc(sizeof *env * env_cap);

				const char *key = lua_tostring(ctx, -2);
				const char *val = lua_tostring(ctx, -1);

				char *entry = malloc(strlen(key) + 1 + strlen(val) + 1);
				snprintf(entry, sizeof entry, "%s=%s", key, val);

				env[env_n++] = entry;
			}

			lua_pop(ctx, 1);
		}
		lua_pop(ctx, 1);
	}

	int stdin_pipe[2] = { -1, -1 };
	int stdout_pipe[2] = { -1, -1 };
	int stderr_pipe[2] = { -1, -1 };

	if (lua_getfield(ctx, 1, "stdin")) {
		if (pipe(stdin_pipe) < 0) {
			int _ = errno;
			close_pipes(stdin_pipe, stdout_pipe, stderr_pipe);
			free_exec(env, env_n);
			errno = _;
			lua_pushnil(ctx);
			lua_pushfstring(ctx, "Failed to create stdin pipe: %s", strerror(errno));
			return 2;
		}
		lua_pop(ctx, 1);
	}
	if (lua_getfield(ctx, 1, "stdout")) {
		if (lua_toboolean(ctx, -1)) {
			if (pipe(stdout_pipe) < 0) {
				int _ = errno;
				close_pipes(stdin_pipe, stdout_pipe, stderr_pipe);
				free_exec(env, env_n);
				errno = _;
				lua_pushnil(ctx);
				lua_pushfstring(ctx, "Failed to create stdout pipe: %s", strerror(errno));
				return 2;
			}
		}

		lua_pop(ctx, 1);
	}
	if (lua_getfield(ctx, 1, "stderr")) {
		if (lua_toboolean(ctx, -1)) {
			if (pipe(stderr_pipe) < 0) {
				int _ = errno;
				close_pipes(stdin_pipe, stdout_pipe, stderr_pipe);
				free_exec(env, env_n);
				errno = _;
				lua_pushnil(ctx);
				lua_pushfstring(ctx, "Failed to create stderr pipe: %s", strerror(errno));
				return 2;
			}
		}

		lua_pop(ctx, 1);
	}

	pid_t child = fork();
	if (child < 0) {
		int _ = errno;
		close_pipes(stdin_pipe, stdout_pipe, stderr_pipe);
		free_exec(env, env_n);
		errno = _;
		lua_pushnil(ctx);
		lua_pushfstring(ctx, "Failed to fork: %s", strerror(errno));
		return 2;
	}
	else if (child == 0) {
		close(stdin_pipe[1]);
		close(stdout_pipe[0]);
		close(stderr_pipe[0]);

		if (stdin_pipe[0] >= 0) dup2(stdin_pipe[0], STDIN_FILENO);
		if (stdout_pipe[1] >= 0) dup2(stdout_pipe[1], STDOUT_FILENO);
		if (stderr_pipe[1] >= 0) dup2(stderr_pipe[1], STDERR_FILENO);

		int res;

		if (env_n == 0) {
			res = execvp(args[0], (char**)args);
		}
		else {
			if (env_n <= env_cap) {
				env = realloc(env, sizeof *env * env_n + 1);
			}
			env[env_n] = NULL;

			res = execvpe(args[0], (char**)args, env);
		}

		// We can't really log much, can we?
		exit(errno);
	}
	else {
		lua_createtable(ctx, 0, 4);

		const char *args[n];

		if (stdin_pipe[1] >= 0) {
			luaL_Stream *stdin_stream = lua_newuserdatauv(ctx, sizeof(luaL_Stream), 0);
			stdin_stream->closef = lib_close;
			stdin_stream->f = fdopen(stdin_pipe[1], "w");
			luaL_getmetatable(ctx, LUA_FILEHANDLE);
			lua_setmetatable(ctx, -2);
			lua_setfield(ctx, -2, "stdin");
		}

		if (stdout_pipe[0] >= 0) {
			luaL_Stream *stdout_stream = lua_newuserdatauv(ctx, sizeof(luaL_Stream), 0);
			stdout_stream->closef = lib_close;
			stdout_stream->f = fdopen(stdout_pipe[0], "r");
			luaL_getmetatable(ctx, LUA_FILEHANDLE);
			lua_setmetatable(ctx, -2);
			lua_setfield(ctx, -2, "stdout");
		}

		if (stderr_pipe[0] >= 0) {
			luaL_Stream *stderr_stream = lua_newuserdatauv(ctx, sizeof(luaL_Stream), 0);
			stderr_stream->closef = lib_close;
			stderr_stream->f = fdopen(stderr_pipe[0], "r");
			luaL_getmetatable(ctx, LUA_FILEHANDLE);
			lua_setmetatable(ctx, -2);
			lua_setfield(ctx, -2, "stderr");
		}

		lua_pushinteger(ctx, child);
		lua_pushcclosure(ctx, lib_reap, 1);
		lua_setfield(ctx, -2, "reap");

		close(stdin_pipe[0]);
		close(stdout_pipe[1]);
		close(stderr_pipe[1]);

		return 1;
	}
}

extern int lualib_open_exec(lua_State *ctx) {
	lua_pushcfunction(ctx, lib_exec);
	return 1;
}
