#define _GNU_SOURCE

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>

static const char *fs_check_path(lua_State *ctx, int i, size_t *psize) {
	size_t n;
	const char *path = luaL_checklstring(ctx, i, &n);

	size_t real_len = strlen(path);
	if (n != real_len) {
		luaL_error(ctx, "path may not contain \\0");
		return NULL;
	}

	if (psize) *psize = n;

	return path;
}

static int readdir_next(lua_State *ctx) {
	DIR *dir = lua_touserdata(ctx, lua_upvalueindex(1));

	while (true) {
		struct dirent *entry = readdir(dir);
		if (!entry) {
			lua_pushnil(ctx);
			return 1;
		}

		if (!strcmp(entry->d_name, ".")) continue;
		if (!strcmp(entry->d_name, "..")) continue;

		lua_pushstring(ctx, entry->d_name);
		return 1;
	}
}

static int lib_fs_mkdir(lua_State *ctx) {
	size_t path_size;
	const char *ro_path = fs_check_path(ctx, 1, &path_size);
	bool recursive = lua_toboolean(ctx, 2);

	if (recursive) {
		char path[path_size + 1];
		memcpy(path, ro_path, path_size + 1);

		for (size_t i = 0; i <= path_size ; i++) {
			if (path[i] == '/' || path[i] == '\0') {
				if (i == 0) continue;

				path[i] = '\0';

				if (mkdir(path, 0755) < 0 && errno != EEXIST) {
					lua_pushnil(ctx);
					lua_pushfstring(ctx, "can't make directory: %s", strerror(errno));
					return 2;
				}

				if (i < path_size) path[i] = '/';
			}
		}
	}
	else {
		if (mkdir(ro_path, 0755) < 0) {
			lua_pushnil(ctx);
			lua_pushfstring(ctx, "can't make directory: %s", strerror(errno));
			return 2;
		}
	}

	lua_pushboolean(ctx, true);
	return 1;
}
static int lib_fs_readdir(lua_State *ctx) {
	const char *path = fs_check_path(ctx, 1, NULL);

	DIR *dir = opendir(path);
	if (!dir) {
		lua_pushnil(ctx);
		lua_pushfstring(ctx, "failed to open directory: %s", strerror(errno));
		return 2;
	}

	lua_pushlightuserdata(ctx, dir);
	lua_pushcclosure(ctx, readdir_next, 1);
	return 1;
}

static int lib_fs_is_file(lua_State *ctx) {
	const char *path = fs_check_path(ctx, 1, NULL);
	struct stat stat;
	if (lstat(path, &stat) < 0) {
		lua_pushnil(ctx);
		lua_pushfstring(ctx, "failed to open directory: %s", strerror(errno));
		return 2;
	}

	lua_pushboolean(ctx, S_ISREG(stat.st_mode));
	return 1;
}
static int lib_fs_is_dir(lua_State *ctx) {
	const char *path = fs_check_path(ctx, 1, NULL);
	struct stat stat;
	if (lstat(path, &stat) < 0) {
		lua_pushnil(ctx);
		lua_pushfstring(ctx, "failed to open directory: %s", strerror(errno));
		return 2;
	}

	lua_pushboolean(ctx, S_ISDIR(stat.st_mode));
	return 1;
}
static int lib_fs_lstat(lua_State *ctx) {
	const char *path = fs_check_path(ctx, 1, NULL);
	struct stat stat;
	if (lstat(path, &stat) < 0) {
		lua_pushnil(ctx);
		lua_pushfstring(ctx, "failed to stat: %s", strerror(errno));
		return 2;
	}

	lua_createtable(ctx, 0, 7);

	lua_pushinteger(ctx, stat.st_mode);
	lua_setfield(ctx, -2, "mode");
	lua_pushinteger(ctx, stat.st_size);
	lua_setfield(ctx, -2, "size");

	lua_pushinteger(ctx, stat.st_uid);
	lua_setfield(ctx, -2, "uid");
	lua_pushinteger(ctx, stat.st_gid);
	lua_setfield(ctx, -2, "gid");

	lua_pushinteger(ctx, stat.st_ctim.tv_sec);
	lua_setfield(ctx, -2, "ctime");
	lua_pushinteger(ctx, stat.st_mtim.tv_sec);
	lua_setfield(ctx, -2, "mtime");
	lua_pushinteger(ctx, stat.st_atim.tv_sec);
	lua_setfield(ctx, -2, "atime");

	return 1;
}

extern int lualib_open_fs(lua_State *ctx) {
	luaL_newlib(ctx, ((luaL_Reg[]) {
		{ "mkdir", lib_fs_mkdir },
		{ "readdir", lib_fs_readdir },

		{ "is_file", lib_fs_is_file },
		{ "is_dir", lib_fs_is_dir },
		{ NULL, NULL },
	}));

	return 1;
}
