local recurse = require "recurse";
local fs = require "fs";

local help_msg = [[
mklua by TopchetoEU

A tool to compile a simple application, consisting of lua and C files, as well
as some assets to one executable.

[file].lua - Shorthand for --lua-file [file].lua
[file].c - Shorthand for --c-file [file].c
--lua-dir [dir] - Specifies a folder, in which to look for lua files
--lua-file [file]
--lua-file [basename]:[file] - Includes a single lua file
--c-dir [dir] - Specifies a folder, in which to look for C files
--c-file [file]
--c-file [basename]:[file] - Includes a single C file
--asset-dir [dir] - Specifies a folder, in which to look for assets
--asset-file [file]
--asset-file [basename]:[file] - Specifies a folder, in which to look for assets
--cc-args - Arguments, which to pass to the C compiler
--entry - The module, which to use as an entry
--output - The file, to which to compile
--debug - Emits debug symbols for lua and C
--dump - Prints out the generated C entry file and exists
--help - shows this message

A mklua "project" consists of modules and assets, which are searched in their
respective directories.

Modules consist of lua and C modules. Lua modules will be searched for in the
directories, specified with "--lua-dir", and will be put in "package.preload"
(the name of the preloaded module is in the lua format, relative to the folder,
in which the file was found). The same goes with the C modules, with the
exception that a C module must declare a lua_CFunction of the name
"lualib_open_path_to_my_module" (for a C module that is called
'path.to.my.module'). That function's return value will be used to initialize
the module's value.

Assets on the other hand will be searched in the same way modules are, but will
be put in a module, called "assets". The value of this module will be a map of
the asset filenames, relative to the folder they were found in, mapped to their
content, as string.

The entry is a module, that is expected to return a function. That function will
be called with the CLI args, when the program is run.
]];


local c_format = [[
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

%s

char bytecode[] = %s;

void load_libraries(lua_State *ctx) {
	luaL_openlibs(ctx);
	%s
}

static int print_error(lua_State *ctx) {
	luaL_traceback(ctx, ctx, "unhandled error: ", 1);
	const char *str = lua_tostring(ctx, -1);
	fprintf(stderr, "%%s\n", str);
	return 0;
}

int main(int argc, const char **argv) {
	lua_State *ctx = luaL_newstate();
	load_libraries(ctx);

	lua_pushcfunction(ctx, print_error);
	int err_handler = lua_gettop(ctx);

	if (luaL_loadbuffer(ctx, bytecode, sizeof bytecode - 1, "<main>")) {
		fprintf(stderr, "lua bytecode error: %%s", lua_tostring(ctx, -1));
		lua_close(ctx);
		return 1;
	}

	for (int i = 1; i < argc; i++) {
		lua_pushstring(ctx, argv[i]);
	}

	if (lua_pcall(ctx, argc - 1, 0, err_handler)) {
		lua_close(ctx);
		return 1;
	}
	else {
		lua_close(ctx);
		return 0;
	}
}
]];

local assets_load_func = [[
static int get_assets(lua_State *ctx) {
	lua_pushvalue(ctx, lua_upvalueindex(1));
	return 1;
}
]];

local asset_init_format = [[
	lua_createtable(ctx, 0, 0);
	lua_pushvalue(ctx, -1);

	luaL_getsubtable(ctx, LUA_REGISTRYINDEX, LUA_LOADED_TABLE);
	lua_pushvalue(ctx, -2);
	lua_setfield(ctx, -2, "assets");
	lua_pop(ctx, 2);
]];

local asset_add_format = [[
	lua_pushlstring(ctx, %s, %d);
	lua_setfield(ctx, -2, %q);
]];

local entry_invoke_format = [[
	return xpcall(require %q, function (err)
		print("unhandled error: " .. tostring(err));

		local i = 2;
		local prefix = "    ";

		while true do
			local info = debug.getinfo(i, "Snl");
			if info == nil then break end

			local name = info.name;
			local location = info.short_src;

			if location == "[C]" then
				location = "at <internal>";
			elseif string.find(location, "[string", 1, true) then
				location = "at <string>";
			else
				location = "at " .. location;
			end

			if info.currentline > 0 then
				location = location .. ":" .. info.currentline;
			end

			if name ~= nil then
				print(prefix .. location .. " in " .. name);
			else
				print(prefix .. location);
			end

			i = i + 1;
		end
	end, ...);
]];

--- @generic T
--- @param arr T[]
--- @return fun(): T?
local function iterate(arr)
	local i = 0;
	local n = #arr;

	return function ()
		if arr == nil then
			return nil;
		end

		i = i + 1;

		if i > n then
			--- @diagnostic disable-next-line: cast-local-type
			arr = nil;
			return nil;
		else
			return arr[i];
		end
	end
end

local function get_module_name(f, prefix)
	return f
		-- Remove prefix slashes
		:gsub("^[/\\]+", "")
		-- Replace slashes with dots
		:gsub("[/\\]", ".")
		-- Remove .init suffix
		:gsub("%.init$", "");
end

local function c_escape(str)
	return "\"" .. str:gsub(".", function (c)
		local b = string.byte(c);

		if c == "\n" then
			return "\\n";
		elseif c == "\\" then
			return "\\\\";
		elseif c == "\"" then
			return "\\\"";
		elseif b >= 32 and b <= 126 then
			return c;
		else
			return ("\\%.3o"):format(b);
		end
	end) .. "\"";
end

return function (...)
	--- @type string[]
	local cc_args = {};

	local lua_files = {};
	local c_files = {};
	local asset_files = {};
	local has_assets = false;
	local strip = true;
	local dump = false;
	local entry;

	local out_path;

	require "args".parser {
		function (arg)
			local path = assert(arg());

			local basename = path:match "^(.+)%.lua$";
			if basename then
				lua_files[basename] = path;
				return;
			end

			basename = path:match "^(.+)%.c$";
			if basename then
				c_files[basename] = path;
				return;
			end

			if path:match "^%-" then
				error("Unknown option '" .. path .. "'");
			else
				error("'" .. path .. "' is neither a .c or a .lua file. Use --asset-file if you want to add an asset");
			end
		end,

		help = function ()
			print(help_msg);
			os.exit();
		end,

		["lua-dir"] = function (arg)
			local dir = assert(arg());
			for path in recurse(dir) do
				local basename = path:match "^(.+)%.lua$";
				if basename and fs.is_file(path) then
					lua_files[basename:sub(#dir + 1)] = path;
				end
			end
		end,
		["lua-file"] = function (arg)
			local val = assert(arg());

			local basename, realname = val:match "^(.+):(.+)$";
			if basename then
				lua_files[basename] = realname;
				return;
			end

			basename = val:match "$(.+)%.[^%.]+";
			if basename then
				lua_files[basename] = val;
				return;
			end

			lua_files[val] = val;
		end,

		["c-dir"] = function (arg)
			local dir = assert(arg());
			for path in recurse(dir) do
				local basename = path:match "^(.+)%.c$";
				if basename and fs.is_file(path) then
					c_files[basename:sub(#dir + 1)] = path;
				end
			end
		end,
		["c-file"] = function (arg)
			local val = assert(arg());

			if not fs.is_file(val) then
				error ""
			end

			local basename, realname = val:match "^(.+):(.+)$";
			if basename then
				c_files[basename] = realname;
				return;
			end

			basename = val:match "$(.+)%.[^%.]+";
			if basename then
				c_files[basename] = val;
				return;
			end

			c_files[val] = val;
		end,

		["asset-dir"] = function (arg)
			local dir = assert(arg());
			for path in recurse(dir) do
				if fs.is_file(path) then
					asset_files[path:sub(#dir + 1)] = path;
				end
			end

			has_assets = true;
		end,
		["asset-file"] = function (arg)
			local val = assert(arg());

			local basename, realname = val:match "^(.+):(.+)$";
			if basename then
				c_files[basename] = realname;
			else
				c_files[val] = val;
			end

			has_assets = true;
		end,

		["cc-arg"] = function (arg)
			table.insert(cc_args, assert(arg()));
		end,

		entry = function (arg)
			entry = assert(arg());
		end,
		debug = function ()
			strip = false;
		end,
		dump = function ()
			dump = true;
		end,

		output = function (arg)
			if out_path then
				error "--output may be specified once";
			end

			out_path = assert(arg());
		end,

		o = "output",
		e = "entry",
		g = "debug",
		h = "help",
	} (...);

	if not out_path then
		error "no output file specified; aborting...";
	end
	if not entry then
		error "no entry specified; aborting...";
	end

	local lib_parts = {};
	local extern_parts = {};

	local lua_parts = {};

	local files = {};

	for basename, path in pairs(lua_files) do
		local module_name = get_module_name(basename);

		-- We do the separate bytecode to keep line numbers and file names
		local bytecode = string.dump(assert(load(io.lines(path, 4096), "@" .. path, "t")), strip);
		table.insert(lua_parts, ("package.preload[%q] = load(%q, %q, 'b');"):format(module_name, bytecode, path));
	end

	table.insert(lua_parts, entry_invoke_format:format(entry));

	if has_assets then
		table.insert(extern_parts, assets_load_func);
		table.insert(lib_parts, asset_init_format);
	end

	for basename, path in pairs(asset_files) do
		local f = assert(io.open(path, "r"));
		local content = assert(f:read "*a");
		assert(f:close());

		table.insert(lib_parts, asset_add_format:format(content, #content, basename));
	end

	for basename, path in pairs(c_files) do
		local module_name = get_module_name(basename);
		local func_name = "lualib_open_" .. module_name:gsub("%.", "_");

		table.insert(extern_parts, ("extern int %s(lua_State *ctx);"):format(func_name));
		table.insert(lib_parts, ("luaL_requiref(ctx, %q, %s, false);"):format(module_name, func_name));
		table.insert(files, path);
	end

	local bytecode = string.dump(assert(load(iterate(lua_parts), "@<entry>", "t")), true);
	local file = c_format:format(table.concat(extern_parts, "\n"), c_escape(bytecode), table.concat(lib_parts, "\n\t"));

	if dump then
		io.stdout:write(file);
		return;
	end

	local f = assert(io.popen(table.concat {
		"cc -llua -x c - ",
		table.concat(files, " "),
		" ",
		strip and "" or "-g ",
		"-o ", out_path, " ",
		table.concat(cc_args, " ")
	}, "w"));

	assert(f:write(file));
	assert(f:close());
end
