local fs = require "fs";
local core = {};

--- @alias task fun(...)

--- @type { [string]: task }
local tasks = {};

--- Runs a task of the given name
--- @param name string
function core.run(name, ...)
	local task = tasks[name];
	if not task then
		error("No task '" .. name "'");
	end

	return task(...);
end

--- Creates a task of the given name
--- @param name string
--- @param task task
function core.task(name, task)
	tasks[name] = task;
end

--- Converts a file1/file2/**/??/*.txt pattern to a function that can match it
--- % is the escape character, and can escape itself
local function mk_pattern(wildcard)
	if type(wildcard) == "function" then
		return wildcard;
	end

	local parts = {};
	local prev = nil;

	for c in wildcard:gmatch "." do
		if prev == "%" then
			table.insert(parts, "%" .. c);
		elseif prev == "*" then
			if c == "*" then
				table.insert(parts, "%" .. ".*");
			else
				table.insert(parts, "%" .. "[^/]*");
				table.insert(parts, c);
			end
		elseif c ~= "*" then
			table.insert(parts, c);
		end

		prev = c;
	end

	if prev == "*" then
		table.insert(parts, "%" .. "[^/]*");
	end

	local pattern = table.concat(parts);

	return function (text)
		return string.match(text, pattern) ~= nil;
	end;
end



--- Returns an iterator of file paths inside the 'root' dir.
--- If 'pattern' is specified, it is going to be tested for each node in the 'root' dir,
--- BEFORE rerooting
--- If absolute_reroot or relative_reroot are specified, the output's absolute and relative paths
--- will be prefixed with the specified directories, AFTER the pattern check has been performed
--- @param settings string | { root: string, pattern?: string }
--- @return fun(): string?
function core.files(settings)
	if type(settings) == "function" then
		return settings;
	elseif type(settings) == "string" then
		return core.files { root = settings };
	end

	local root = settings.root or error "no root specified";
	local pattern = settings.pattern and mk_pattern(settings.pattern);

	local recurser, recurse_state = require "recurse" (root);

	return function ()
		for el in recurser, recurse_state do
			local relative = el:sub(#root + 1);
			if not pattern or pattern(relative) then
				return root .. relative;
			end
		end

		return nil;
	end
end

return core;

-- --- @diagnostic disable: cast-local-type

-- --- @alias pattern fun(text: string): boolean

-- local exec = require "exec";
-- local fs = require "fs";

-- local core = {};

-- --- Converts a file1/file2/**/??/*.txt pattern to a function that can match it
-- --- % is the escape character, and can escape itself
-- function core.pattern(wildcard)
-- 	if type(wildcard) == "function" then
-- 		return wildcard;
-- 	end

-- 	local parts = {};
-- 	local prev = nil;

-- 	for c in wildcard:gmatch "." do
-- 		if prev == "%" then
-- 			table.insert(parts, "%" .. c);
-- 		elseif prev == "*" then
-- 			if c == "*" then
-- 				table.insert(parts, "%" .. ".*");
-- 			else
-- 				table.insert(parts, "%" .. "[^/]*");
-- 				table.insert(parts, c);
-- 			end
-- 		elseif c ~= "*" then
-- 			table.insert(parts, c);
-- 		end

-- 		prev = c;
-- 	end

-- 	if prev == "*" then
-- 		table.insert(parts, "%" .. "[^/]*");
-- 	end

-- 	local pattern = table.concat(parts);

-- 	return function (text)
-- 		return string.match(text, pattern) ~= nil;
-- 	end;
-- end


-- function core.combine(arr)
-- 	local arr_i = 1;
-- 	local curr = arr[1];

-- 	return function ()
-- 		while true do
-- 			if arr == nil or arr_i > #arr then
-- 				arr = nil;
-- 				return nil;
-- 			end

-- 			local res = curr();
-- 			if res ~= nil then return res end

-- 			arr_i = arr_i + 1;
-- 			curr = arr[arr_i];
-- 		end
-- 	end
-- end

-- --- Creates a generic transformer that independently converts each file to one or more files
-- --- The passed function is called once with a settings object. The function it
-- --- returns will be called upon each transformation
-- function core.create_transformer(func, all)
-- 	local function finish_call(func, ...)
-- 		local res = func(...);
-- 		if type(res) == "string" then
-- 			res = { res };
-- 		end
-- 		if type(res) == "table" then
-- 			local i = 0;
-- 			local arr = res;
-- 			res = function ()
-- 				if not arr then return nil end
-- 				i = i + 1;
-- 				if i > #arr then return nil end

-- 				return arr[i];
-- 			end
-- 		end

-- 		if type(res) == "function" then
-- 			return res;
-- 		else
-- 			error "Invalid return from transformer function";
-- 		end
-- 	end

-- 	return function (settings)
-- 		local input = core.files(settings.input);
-- 		local output = settings.output or error "no output specified";
-- 		local res_func = func(settings);

-- 		if all then
-- 			return function ()
-- 				if not input then return nil end

-- 				local a, b = input, output;
-- 				input, output = nil, nil;

-- 				return finish_call(res_func, a, b);
-- 			end
-- 		else
-- 			return function ()
-- 				if not input then return nil end

-- 				local curr = input();
-- 				if not curr then
-- 					input, output = nil, nil;
-- 					return nil;
-- 				else
-- 					return finish_call(res_func, input, output);
-- 				end
-- 			end
-- 		end
-- 	end
-- end

-- --- A generic cc a.o b.o c.o -o res adapter
-- core.obj = core.create_transformer(function (settings)
-- 	local cc = settings.cc or "cc";

-- 	return function (input, output)
-- 		local rest = {};
-- 		for el in input do table.insert(rest, el) end
-- 		table.move(settings.flags, 1, #settings.flags, #rest + 1, rest);

-- 		simple_exec(cc, "-o", output, table.unpack(rest));
-- 	end
-- end, true);
-- --- A generic luac adapter
-- core.lua = core.create_transformer(function (settings)
-- 	local lua = settings.lua or "lua";

-- 	return function (input, output)
-- 		simple_exec(lua,
-- 			input.abs,
-- 			"-o", output.abs .. "/" .. input.rel .. ".o",
-- 			table.unpack(settings.flags)
-- 		);
-- 	end
-- end);
-- --- A generic luac adapter
-- core.lua_bundle = core.create_transformer(function (settings)
-- 	local entries = {};

-- 	if settings.entry then
-- 		table.insert(entries, settings.entry);
-- 	end

-- 	if settings.entries then
-- 		table.move(settings.entries, 1, #settings.entries, #entries + 1, entries);
-- 	end

-- 	if #entries == 0 then
-- 		error "No entries have been specified";
-- 	end

-- 	return function (input, output)
-- 		local f = assert(io.open(output.abs, "w"));

-- 		for path in input do
-- 			local mod_name = get_module_name(path.rel);
-- 			local bytecode = string.dump(assert(load(io.lines(path.abs, 4096), "@" .. path.abs, "t")), settings.strip);
-- 			f:write(("package.preload[%q] = load(%q, nil, \"b\");"):format(mod_name, bytecode));
-- 		end

-- 		for entry in entries do
-- 			f:write(("require %q (...);"):format(entry));
-- 		end
-- 	end
-- end, true);
-- --- Generates a C file, containing a binary blob
-- core.c_blob = core.create_transformer(function (settings)
-- 	return function (input, output)
-- 		local f = assert(io.open(output.abs, "w"));

-- 		f:write("unsigned char %s[] = ");
-- 		f:write(c_escape())

-- 		for path in input do
-- 			local mod_name = get_module_name(path.rel);
-- 			local bytecode = string.dump(assert(load(io.lines(path.abs, 4096), "@" .. path.abs, "t")), settings.strip);
-- 			f:write(("package.preload[%q] = load(%q, nil, \"b\");"):format(mod_name, bytecode));
-- 		end
-- 	end
-- end);
-- --- Generates a lua file, containing a binary blob
-- core.lua_blob = core.create_transformer(function (settings)
-- 	return function (input, output)
-- 	end
-- end);

-- return core;
