local utils = require "utils";
local lua = {};

local function get_module_name(path)
	return path
		-- Remove extension
		:gsub("%.[^%.]-$", "")
		-- Remove prefix slashes
		:gsub("^[/\\]+", "")
		-- Replace slashes with dots
		:gsub("[/\\]", ".")
		-- Remove .init suffix
		:gsub("%.init$", "");
end

local function compile(input, output, filename, strip, exe)
	filename = filename or input;

	if not exe then
		local func = assert(load(assert(io.lines(input, 4096)), filename, "t"));
		local bytecode = string.dump(func, strip);

		local out_f = assert(io.open(output, "w"));
		assert(out_f:write(bytecode));
		assert(out_f:close());
	else
		utils.simple_exec(exe, "-e", ([[
			local func = assert(load(assert(io.lines(%q, 4096)), %q, "t"));
			local bytecode = string.dump(func, %s);
			local out_f = assert(io.open(%q, "w"));
			assert(out_f:write(bytecode));
			assert(out_f:close());
		]]):format(input, filename, strip and "true" or "false", output));
	end
end

--- Compiles the given lua file to bytecode
--- @param settings { input: string, output: string, filename?: string, strip?: boolean, exe?: string }
--- @return task
function lua.compile(settings)
	local input = settings.input;
	local output = settings.output;
	local filename = "@" .. (settings.filename or settings.input);
	local strip = settings.strip or false;
	local exe = settings.exe;

	return function ()
		if utils.check_files({ input }, { output }) then return end
		compile(input, output, filename, strip, exe);
	end
end

--- Bundles multiple lua files into a single one
--- @param settings { inputs: string[] | (fun(): string?), relative_to?: string, output: string, entry: string }
--- @return task
function lua.bundle(settings)
	local inputs = utils.to_array(settings.inputs);
	local output = settings.output;
	local entry = settings.entry;
	local relative_to = settings.relative_to or "";

	return function ()
		local out_f = assert(io.open(output, "w"));

		for i = 1, #inputs do
			local path = inputs[i]:sub(#relative_to);
			local name = get_module_name(path);

			local in_f = assert(io.open(inputs[i], "r"));
			local data = assert(in_f:read "*a");
			in_f:close();

			assert(out_f:write(("package.preload[%q] = load(%q, \"bt\", %q);\n"):format(name, data, path)));
		end

		assert(out_f:write(("require %q (...);\n"):format(entry)));
		out_f:close();
	end
end

--- Generates a single C file that exposes the given C libraries to lua
--- @param settings { inputs: string[] | (fun(): string?), relative_to?: string, output: string, func_name?: string }
--- @return task
function lua.c_libs(settings)
	local inputs = utils.to_array(settings.inputs);
	local output = settings.output;
	local func_name = settings.func_name or "mklua_init_libs";
	local relative_to = settings.relative_to or "";

	return function ()
		local out_f = assert(io.open(output, "w"));

		assert(out_f:write("#include <lua.h>\n"));
		assert(out_f:write("#include <lualib.h>\n"));
		assert(out_f:write("#include <lauxlib.h>\n"));

		assert(out_f:write("\n"));

		for i = 1, #inputs do
			local path = inputs[i]:sub(#relative_to);
			local name = get_module_name(path);

			local in_f = assert(io.open(inputs[i], "r"));
			local data = assert(in_f:read "*a");
			in_f:close();

			assert(out_f:write(("package.preload[%q] = load(%q, \"bt\", %q);\n"):format(name, data, path)));
		end

		assert(out_f:write(("require %q (...);\n"):format("")));
		out_f:close();
	end
end

return lua;
