local exec = require "exec";
local fs = require "fs";

local utils = {};

utils.exec = exec;

function utils.simple_exec(program, ...)
	local child = assert(exec { program, ... });

	child.stdin:close(io.stdin);
	child.stdout:close(io.stdout);
	child.stderr:close(io.stderr);

	local code = assert(child.reap());
	if code ~= 0 then
		error(("'%s' failed with exit code %d"):format(program, code));
	end
end
function utils.ensure_dir(path)
	local dirname = path:match "^(.+)/[^/]-$";
	if dirname then
		assert(fs.mkdir(dirname, true));
	end
end

--- @generic T
--- @param val T[] | (fun(): T?)
function utils.iterate(val)
	if type(val) == "function" then
		return val;
	else
		local i = 0;
		return function ()
			if not val then
				return nil;
			end

			i = i + 1;
			if i > #val then
				--- @diagnostic disable-next-line: cast-local-type
				val = nil;
				return nil;
			end

			return val[i];
		end
	end
end

--- @generic T
--- @param val T[] | (fun(): T?)
--- @return T[]
function utils.to_array(val)
	local res = {};
	for el in utils.iterate(val) do
		table.insert(res, el);
	end
	return res;
end

--- Checks if all the output files are present, and if input files are
--- specified, whether or not they are older than all the output files.
---
--- Returns true if all checks hold
---
--- This function can be used to prevent redundant builds
---
--- @param outputs string[] | (fun(): string?)
--- @param inputs? string[] | (fun(): string?)
--- @return boolean
function utils.check_files(inputs, outputs)
	local in_n = 0;
	local out_n = 0;
	local in_mtime, out_mtime;
	local out_missing = false;

	if inputs then
		for el in utils.iterate(inputs) do
			local stat = fs.lstat(el);
			if not stat then error("Input '" .. el "' not present") end

			in_n = in_n + 1;
			if not in_mtime or in_mtime < stat.mtime then
				in_mtime = stat.mtime;
			end
		end
	end

	for el in utils.iterate(outputs) do
		out_n = out_n + 1;

		local stat = fs.lstat(el);
		if not stat then
			out_missing = true;
		elseif not out_mtime or out_mtime > stat.mtime then
			out_mtime = stat.mtime;
		end
	end

	if out_missing then
		return false;
	elseif in_mtime and out_mtime then
		return in_mtime < out_mtime;
	else
		return true;
	end
end

return utils;
